/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.CollationKey;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.wordsource.WordSourceId;

/**
 * Defines utilities used in this component across the packages. Since this is a multi-package component, the public
 * helper class is allowed.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public final class IndexPersistenceHelper {
    /**
     * Represents the digest algorithm used to create the string ID.
     */
    private static final String DIGEST = "MD5";

    /**
     * Represents the bit-wise mask used to obtain one unsigned byte.
     */
    private static final int HEX_MASK = 0xff;

    /**
     * Creates an instance of <code>IndexPersistenceHelper</code> class. This private constructor prevents the
     * creation of a new instance.
     */
    private IndexPersistenceHelper() {
    }

    /**
     * Generates a string ID according to the given document ID. Though the ID is not unique, the probablity of two
     * different document IDs having the same string ID should be extremely low. The string ID is a hex string of the
     * digested key elements in the document ID. MD5 digest algorithm is used. Though the test shows that it is slightly
     * slower than XORing all bytes, it significantly reduces the chances of collision.
     *
     * @param id the document ID whose string ID is generated.
     * @return the string ID generated from the document ID.
     * @throws IllegalArgumentException if <code>id</code> is <code>null</code>.
     * @throws IndexPersistenceException if the source of the document is not serializable.
     */
    public static String generateId(WordSourceId id) throws IndexPersistenceException {
        if (id == null) {
            throw new IllegalArgumentException("id cannot be null.");
        }

        try {
            // Create a digest output stream based on a mock stream
            DigestOutputStream dos = new DigestOutputStream(new OutputStream() {
                /**
                 * Writes the given bytes to the output stream. It does not write any thing.
                 *
                 * @param b the bytes to be written.
                 */
                public void write(byte[] b) {
                }

                /**
                 * Writes the given bytes to the output stream. It does not write any thing.
                 *
                 * @param b the byte array containing the bytes to be written.
                 * @param off the offset of the bytes in the array to be written.
                 * @param len the length of the bytes to be written.
                 */
                public void write(byte b[], int off, int len) {
                }

                /**
                 * Writes the given byte to the output stream. It does not write any thing.
                 *
                 * @param b the byte to be written.
                 */
                public void write(int b) {
                }
            }, MessageDigest.getInstance(DIGEST));

            // Create an object output stream based on the digest stream
            ObjectOutputStream oos = new ObjectOutputStream(dos);

            // Write out the key elements in the document ID to the object output stream
            oos.writeUTF(id.getSourceClassName());
            oos.writeUTF(id.getSourceLocale().toString());
            oos.writeObject(id.getDelimiters());
            oos.writeObject(id.getSourceIdentity());

            // Get the digest bytes
            byte[] digest = dos.getMessageDigest().digest();
            StringBuffer buffer = new StringBuffer(digest.length * 2);

            // Build the hex string of the digest.
            for (int i = 0; i < digest.length; ++i) {
                // Each byte is converted into a two-char hex string
                String s = Integer.toHexString(((int) digest[i]) & HEX_MASK);

                if (s.length() == 1) {
                    buffer.append('0');
                }

                buffer.append(s);
            }

            return buffer.toString();
        } catch (IOException e) {
            throw new IndexPersistenceException("The document source must be serializable.", e);
        } catch (NoSuchAlgorithmException e) {
            throw new IndexPersistenceException("The digest algorithm is not supported.", e);
        }
    }

    /**
     * Adds the document index to the collection word map. The keys of the map are collation keys of all different
     * words. The values are sets of document IDs repesenting the documents containing the words.
     *
     * @param wordMap the collection word map to be added to.
     * @param index the document index to be added to the word map.
     */
    public static void addDocumentWordMap(Map wordMap, DocumentIndex index) {
        for (Iterator iter = index.getWords().keySet().iterator(); iter.hasNext();) {
            CollationKey key = (CollationKey) iter.next();
            Set wordIds = (Set) wordMap.get(key);

            if (wordIds == null) {
                // The word does not exist, create a new set
                wordIds = new HashSet();
                wordMap.put(key, wordIds);
            }

            // Mark the word's existence in the document
            wordIds.add(index.getWordSourceId());
        }
    }
}











