/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.reflect.InvocationTargetException;
import java.sql.Blob;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.Format;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.topcoder.db.connectionfactory.DBConnectionException;
import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;
import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.util.config.ConfigManager;
import com.topcoder.util.config.ConfigManagerException;

/**
 * Defines an abstract implementation of <code>IndexPersistence</code> interface to provide the common functionality
 * of index persistences based on database. It uses the DB connection factory to access the database. The common
 * functionality includes creating new connections from the DB connection factory, and starting, committing or rolling
 * back a database transaction.<br>
 * Besides common functionality of database connection manipulation, some helper functions are also provided for the
 * common operations of the concrete implementations.<br>
 * The class can be configured by the configuration manager. Under the given namespace, three properties may exist,
 * which are 'ConnectionName', 'ConnectionFactoryClassName' and 'ConnectionFactoryNamespace'. Optional property
 * 'ConnectionName' defines the name of the connection used to create new database connections in the DB connection
 * factory. When it is missing or is empty string, the default connection name in the DB connection factory is used.
 * Optional property 'ConnectionFactoryClassName' defines the class name of the DB connection factory implementation
 * used to create new connections. When it is missing,
 * <code>com.topcoder.db.connectionfactory.DBConnectionFactoryImpl</code> is used. Mandatory property
 * 'ConnectionFactoryNamespace' defines the namespace used to create the DB connection factory instance.<br>
 * The thread-safety of this class depends on the DB connection factory instance.
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 */
public abstract class AbstractDBIndexPersistence implements IndexPersistence {
    /**
     * Represents the property name where the connection name is loaded.
     */
    private static final String CONNECTION_NAME_PROPERTY = "ConnectionName";

    /**
     * Represents the property name where the class name of the DB connection factory is loaded.
     */
    private static final String CONNECTION_FACTORY_CLASSNAME_PROPERTY = "ConnectionFactoryClassName";

    /**
     * Represents the property name where the namespace of the DB connection factory is loaded.
     */
    private static final String CONNECTION_FACTORY_NAMESPACE_PROPERTY = "ConnectionFactoryNamespace";

    /**
     * Represents the SQL statement template used to remove a document index from the persistence.
     */
    private static final Format REMOVE_DOCUMENT_INDEX = new MessageFormat("DELETE FROM {0} WHERE DOCUMENT_ID = ?");

    /**
     * Represents the SQL statement used to add a collection index ID to the persistence.
     */
    private static final String ADD_COLLECTION_INDEX_ID = "INSERT INTO DOCUMENT_COLLECTION (COLLECTION_ID) VALUES (?)";

    /**
     * Represents the SQL statement used to add a document index to a collection index in the persistence.
     */
    private static final String ADD_COLLECTION_INDEX_DOCUMENT = "INSERT INTO DOCUMENT_COLLECTION_DOCUMENT_XREF"
        + " (COLLECTION_ID, DOCUMENT_ID) VALUES (?, ?)";

    /**
     * Represents the SQL statement used to remove a collection index from the persistence.
     */
    private static final String DELETE_COLLECTION_INDEX_ID = "DELETE FROM DOCUMENT_COLLECTION WHERE COLLECTION_ID = ?";

    /**
     * Represents the SQL statement used to remove all document indices from a collection index in the persistence.
     */
    private static final String DELETE_COLLECTION_INDEX_DOCUMENT = "DELETE FROM DOCUMENT_COLLECTION_DOCUMENT_XREF"
        + " WHERE COLLECTION_ID = ?";

    /**
     * Represents the SQL statement used to select a collection index from the persistence.
     */
    private static final String SELECT_COLLECTION_INDEX_ID = "SELECT * FROM DOCUMENT_COLLECTION"
        + " WHERE COLLECTION_ID = ?";

    /**
     * Represents the SQL statement used to increase the use count of a document index in the persistence.
     */
    private static final String INCREASE_DOCUMENT_USE_COUNT = "UPDATE DOCUMENT SET USE_COUNT = USE_COUNT + 1"
        + " WHERE DOCUMENT_ID = ? AND USE_COUNT >= 0";

    /**
     * Represents the SQL statement used to decrease the use count of a document index in the persistence.
     */
    private static final String DECREASE_DOCUMENT_USE_COUNT = "UPDATE DOCUMENT SET USE_COUNT = USE_COUNT - 1"
        + " WHERE DOCUMENT_ID = ? AND USE_COUNT > 0";

    /**
     * Represents the SQL statement used to select the use count of a document index in the persistence.
     */
    private static final String SELECT_DOCUMENT_USE_COUNT = "SELECT USE_COUNT FROM DOCUMENT WHERE DOCUMENT_ID = ?";

    /**
     * Represents the SQL statement used to select all document indices in the persistence.
     */
    private static final String SELECT_ALL_DOCUMENT = "SELECT * FROM DOCUMENT";

    /**
     * Represents the SQL statement used to select all delimiters of a document index in the persistence.
     */
    private static final String SELECT_DOCUMENT_DELIMITER = "SELECT DELIMITER_ENTRY FROM DELIMITER"
        + " WHERE DOCUMENT_ID = ? ORDER BY DELIMITER_SORT_ORDER ASC";

    /**
     * Represents the column name where the string ID of the document is stored.
     */
    private static final String DOCUMENT_ID_COLUMN = "DOCUMENT_ID";

    /**
     * Represents the column name where the locale of the document is stored.
     */
    private static final String LOCALE_COLUMN = "LOCALE";

    /**
     * Represents the column name where the document source of the document is stored.
     */
    private static final String SOURCE_COLUMN = "SOURCE_IDENTITY";

    /**
     * Represents the column name where the class name of the document is stored.
     */
    private static final String CLASS_NAME_COLUMN = "CLASS_NAME";

    /**
     * Represents the delimiter used to separate different parts in the locale string.
     */
    private static final String LOCALE_DELIMITER = "_";

    /**
     * Represents an empty string.
     */
    private static final String EMPTY = "";

    /**
     * Represents the SQL statement used to add a document ID into the persistence.
     */
    private static final String ADD_DOCUMENT_INDEX_DOCUMENT = "INSERT INTO DOCUMENT"
        + " (DOCUMENT_ID, LOCALE, CLASS_NAME, SOURCE_IDENTITY, USE_COUNT) VALUES (?, ?, ?, ?, 0)";

    /**
     * Represents the SQL statement used to select a document source ID from the persistence.
     */
    private static final String SELECT_DOCUMENT_INDEX_SOURCE = "SELECT SOURCE_IDENTITY FROM DOCUMENT"
        + " WHERE DOCUMENT_ID = ?";

    /**
     * Represents the SQL statement used to update a document source ID into the persistence.
     */
    private static final String UPDATE_DOCUMENT_INDEX_DOCUMENT = "UPDATE DOCUMENT SET SOURCE_IDENTITY = ?"
        + " WHERE DOCUMENT_ID = ?";

    /**
     * Represents the SQL statement used to add the delimiters of a document index into the persistence.
     */
    private static final String ADD_DOCUMENT_INDEX_DELIMITER = "INSERT INTO DELIMITER"
        + " (DOCUMENT_ID, DELIMITER_ENTRY, DELIMITER_SORT_ORDER) VALUES (?, ?, ?)";

    /**
     * Represents the SQL statement used to select the document IDs of a collection index from the persistence.
     */
    private static final String SELECT_DOCUMENT_INDEX_COLLECTION = "SELECT DOCUMENT.LOCALE, DOCUMENT.SOURCE_IDENTITY,"
        + " DOCUMENT.CLASS_NAME, DOCUMENT.DOCUMENT_ID FROM DOCUMENT INNER JOIN DOCUMENT_COLLECTION_DOCUMENT_XREF"
        + " ON DOCUMENT.DOCUMENT_ID = DOCUMENT_COLLECTION_DOCUMENT_XREF.DOCUMENT_ID"
        + " WHERE DOCUMENT_COLLECTION_DOCUMENT_XREF.COLLECTION_ID = ?";

    /**
     * Represents the DB connection factory used to create database connections.
     */
    private final DBConnectionFactory connectionFactory;

    /**
     * Represents the connection name used to create database connections from the DB connection factory. When it is
     * <code>null</code>, the default connection name is used.
     */
    private final String connectionName;

    /**
     * Creates a new instance of <code>AbstractDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are loaded from the given namespace in the configuration.
     *
     * @param namespace the namespace where the configurations are loaded.
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if the configurations are not valid; or accessing the configuration
     *             manager fails.
     */
    protected AbstractDBIndexPersistence(String namespace) throws PersistenceConfigurationException {
        DBPersistenceHelper.validateNotNullOrEmpty(namespace, "namespace");

        // Load the connection name, DB factory class name and DB factory namespace
        String className = getConfigStringOptional(namespace, CONNECTION_FACTORY_CLASSNAME_PROPERTY,
            DBConnectionFactoryImpl.class.getName());
        String factoryNamespace = getConfigStringOptional(namespace, CONNECTION_FACTORY_NAMESPACE_PROPERTY, null);

        connectionName = getDBConnectionName(getConfigStringOptional(namespace, CONNECTION_NAME_PROPERTY, null));

        // Check DB factory namespace
        if (factoryNamespace == null) {
            throw new PersistenceConfigurationException("The DB connection factory namespace must exist.");
        }

        // Get the class of the DB factory. Note that all Errors should *not* be catched.
        Class factoryClass;

        try {
            factoryClass = Class.forName(className);
        } catch (ClassNotFoundException e) {
            throw new PersistenceConfigurationException("The DB connection factory class cannot be found.", e);
        }

        // Create the DB factory instance.
        try {
            // When the factory namespace is null, use default constructor.
            connectionFactory = (DBConnectionFactory) factoryClass.getConstructor(new Class[] {String.class})
                .newInstance(new Object[] {factoryNamespace});
        } catch (ClassCastException e) {
            throw new PersistenceConfigurationException("The class is not a DB connection factory.", e);
        } catch (IllegalAccessException e) {
            throw new PersistenceConfigurationException(
                "The proper constructor of the DB connection factory cannot be accessed.", e);
        } catch (InstantiationException e) {
            throw new PersistenceConfigurationException("The DB connection factory is not a concrete class.", e);
        } catch (InvocationTargetException e) {
            throw new PersistenceConfigurationException("Constructor of the DB connection factory causes error.", e);
        } catch (NoSuchMethodException e) {
            throw new PersistenceConfigurationException(
                "A property constructor of the DB connection factory cannot be found.", e);
        } catch (SecurityException e) {
            throw new PersistenceConfigurationException("The access to the DB connection factory is denied.", e);
        }
    }

    /**
     * Creates a new instance of <code>AbstractDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are given. When the connection name is empty string or <code>null</code>, the default
     * connection name is used.
     *
     * @param connectionFactory the DB connection factory used to create database connections.
     * @param connectionName the connection name used to create database connections.
     * @throws IllegalArgumentException if <code>connectionFactory</code> is <code>null</code>.
     */
    protected AbstractDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        DBPersistenceHelper.validateNotNull(connectionFactory, "connectionFactory");

        this.connectionFactory = connectionFactory;
        this.connectionName = getDBConnectionName(connectionName);
    }

    /**
     * Gets the connection name used to create database connections. When the given connection name is <code>null</code>
     * or empty string, the default connection name is used.
     *
     * @param connectionName the connection name in DB factory used to create connections.
     * @return <code>null</code> if the default connection name should be used; otherwise <code>connectionName</code>.
     */
    private String getDBConnectionName(String connectionName) {
        // When the connection name is empty string, return null to use default connection name
        return ((connectionName == null) || (connectionName.trim().length() == 0)) ? null : connectionName;
    }

    /**
     * Gets a new database connection using the DB connection factory and the connection name. The retrieved connection
     * is not auto-committed.
     *
     * @return a new database connection created by the DB connection factory.
     * @throws IndexPersistenceException If creating the database connection fails.
     */
    protected Connection getConnection() throws IndexPersistenceException {
        Connection conn;

        try {
            // Create a connection
            // When the connection name is null (empty strings are set as null in the constructors), use default name
            conn = (connectionName == null) ? connectionFactory.createConnection() : connectionFactory
                .createConnection(connectionName);
        } catch (DBConnectionException e) {
            throw new IndexPersistenceException("The database connection cannot be created.", e);
        }

        try {
            // Set the connection as manually committed.
            conn.setAutoCommit(false);
        } catch (SQLException e) {
            try {
                // When the connection does not support transactions, the connection should be closed.
                conn.close();
            } catch (SQLException ignore) {
                // Ignore
            }

            throw new IndexPersistenceException("The connection cannot be set to be manually committed.", e);
        }

        return conn;
    }

    /**
     * Begins a database transaction on the given connection. Currently no operations are done, since the connection
     * should be manually committed.
     *
     * @param connection the connection whose transaction should be started.
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>.
     * @throws IndexPersistenceException If starting the database transaction fails; or the connection does not support
     *             transactions.
     */
    protected void beginTransaction(Connection connection) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(connection, "connection");

        // Do nothing, since the connection is already set as manually committed
        // Reserved for future sub-classes
    }

    /**
     * Commits a database transaction on the given connection. The database connection is closed even if there is an
     * error.
     *
     * @param connection the connection whose transaction should be started.
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>.
     * @throws IndexPersistenceException If committing the database transaction fails; or the connection does not
     *             support transactions.
     */
    protected void commitTransaction(Connection connection) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(connection, "connection");

        try {
            try {
                // Commit the transaction
                connection.commit();
            } finally {
                // No matter what, the connection is closed.
                connection.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("The active transaction cannot be committed.", e);
        }
    }

    /**
     * Rolls back a database transaction on the given connection. The database connection is closed even if there is an
     * error.
     *
     * @param connection the connection whose transaction should be started.
     * @throws IllegalArgumentException if <code>connection</code> is <code>null</code>.
     * @throws IndexPersistenceException If rolling back the database transaction fails; or the connection does not
     *             support transactions.
     */
    protected void rollbackTransaction(Connection connection) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(connection, "connection");

        try {
            try {
                // Roll back the transaction
                connection.rollback();
            } finally {
                // No matter what, the connection is closed.
                connection.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("The active transaction cannot be rolled back.", e);
        }
    }

    /**
     * Gets the property value in the configuration manager. If the property does not exist, the default value is
     * returned.
     *
     * @param namespace the namespace where the property exists.
     * @param propertyName the property name whose value is retrieved.
     * @param defaultValue the default value to be returned when the property is missing and optional.
     * @return the property value when the property exists; or the default value when the property is missing.
     * @throws PersistenceConfigurationException if accessing the configuration manager fails.
     */
    private static String getConfigStringOptional(String namespace, String propertyName, String defaultValue)
        throws PersistenceConfigurationException {
        String value;

        try {
            // Load the value as string
            value = ConfigManager.getInstance().getString(namespace, propertyName);
        } catch (ConfigManagerException e) {
            throw new PersistenceConfigurationException("The configuration value cannot be read.", e);
        }

        // When the property is missing, return the default value
        // If the property exists, return its value
        return (value == null) ? defaultValue : value;
    }

    /**
     * Removes the given document index from the database. All records of the document index in the given tables are
     * deleted from the database. When the database operation fails, the modifications are rolled back.
     *
     * @param sourceId the document ID whose index is deleted from the database.
     * @param tables the array of table names whose records are deleted.
     * @throws IllegalArgumentException if <code>sourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *             or there is at least one collection containing the document index.
     */
    void removeDocumentIndexImpl(WordSourceId sourceId, String[] tables) throws IndexPersistenceException {
        if (getDocumentUseCount(sourceId) != 0) {
            // If the usage count is not 0, throw exception
            throw new IndexPersistenceException("The document index usage count is not 0.");
        }

        // Generate the string ID of the document
        String id = IndexPersistenceHelper.generateId(sourceId);

        // Obtain the connection and start the transaction
        Connection conn = getConnection();

        beginTransaction(conn);

        try {
            String[] params = new String[1];

            // Delete all document index records in the given tables
            for (int tableIndex = 0; tableIndex < tables.length; ++tableIndex) {
                params[0] = tables[tableIndex];
                DBPersistenceHelper.executeSQL(conn, REMOVE_DOCUMENT_INDEX.format(params), new String[] {id}).close();
            }
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Error occurs when deleting the document index.", e);
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails. When
     * the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be persisted.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in the
     *             persistence.
     */
    void addCollectionIndexImpl(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        // Obtain the database connection and start the transaction
        Connection conn = getConnection();

        beginTransaction(conn);

        try {
            // Add the collection ID to the database. When the collection exists, the database would cause error.
            DBPersistenceHelper.executeSQL(conn, ADD_COLLECTION_INDEX_ID, new String[] {collectionIndex.getId()})
                .close();

            // Add all document indices
            addDocumentsToCollection(collectionIndex, conn);
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Error occurs when persisting the collection index.", e);
        } catch (IndexPersistenceException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw e;
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Adds the document indices in the collection index to the database.
     *
     * @param collectionIndex the collection index whose document indices are persisted.
     * @param conn the database connection used to persist the document indices.
     * @throws SQLException if error occurs when persisting the document indices in the collection index.
     * @throws IndexPersistenceException if a string ID cannot be generated from the document ID.
     */
    private static void addDocumentsToCollection(CollectionIndex collectionIndex, Connection conn) throws SQLException,
        IndexPersistenceException {
        // Add each document index in the collection
        PreparedStatement statement = conn.prepareStatement(ADD_COLLECTION_INDEX_DOCUMENT);

        try {
            // Set the collection ID only once
            statement.setString(1, collectionIndex.getId());

            for (Iterator iter = collectionIndex.getAllDocumentIds().iterator(); iter.hasNext();) {
                // Set the string ID of the document index
                statement.setString(2, IndexPersistenceHelper.generateId((WordSourceId) iter.next()));

                // Update the database
                statement.executeUpdate();
            }
        } finally {
            statement.close();
        }
    }

    /**
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionId the ID of the collection index to be deleted.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    void removeCollectionIndexImpl(String collectionId) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNullOrEmpty(collectionId, "collectionId");

        // Obtain the database connection and start the transaction
        Connection conn = getConnection();

        beginTransaction(conn);

        try {
            // Remove the collection ID from the database.
            PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, DELETE_COLLECTION_INDEX_ID,
                new String[] {collectionId});
            int records;

            try {
                records = statement.getUpdateCount();
            } finally {
                statement.close();
            }

            if (records == 0) {
                // When the collection cannot be found
                throw new SQLException("The collection cannot be found in the persistence.");
            }

            removeDocumentsFromCollection(collectionId, conn);
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Error occurs when deleting the collection index.", e);
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Removes the document indices from the collection index in the database.
     *
     * @param collectionId the collection ID whose document indices are deleted.
     * @param conn the database connection used to delete the document indices.
     * @throws SQLException if error occurs when deleting the document indices from the collection index.
     */
    private static void removeDocumentsFromCollection(String collectionId, Connection conn) throws SQLException {
        // Delete all document indices in the collection
        DBPersistenceHelper.executeSQL(conn, DELETE_COLLECTION_INDEX_DOCUMENT, new String[] {collectionId}).close();
    }

    /**
     * Checks if the collection index exists in the persistence. The ID of the collection index is given.
     *
     * @param collectionId the ID of the collection index to be checked.
     * @param conn the database connection used to access the persistence.
     * @return <code>true</code> if the collection index exists; <code>false</code> otherwise.
     * @throws SQLException if accessing the database fails.
     */
    private static boolean existCollection(String collectionId, Connection conn) throws SQLException {
        PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_COLLECTION_INDEX_ID,
            new String[] {collectionId});
        boolean exist;

        try {
            exist = statement.getResultSet().next();
        } finally {
            statement.close();
        }

        return exist;
    }

    /**
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the persistence,
     * it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be updated.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    void updateCollectionIndexImpl(CollectionIndex collectionIndex) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(collectionIndex, "collectionIndex");

        // Obtain the database connection and start the transaction
        Connection conn = getConnection();

        beginTransaction(conn);

        try {
            String collectionId = collectionIndex.getId();

            if (!existCollection(collectionId, conn)) {
                // When the collection index does not exist, throw exception
                throw new SQLException("The collection does not exist in the persistence.");
            }

            // Updating is the same as first removing all and then adding all.
            removeDocumentsFromCollection(collectionId, conn);
            addDocumentsToCollection(collectionIndex, conn);
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Error occurs when updating the collection index.", e);
        } catch (IndexPersistenceException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw e;
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     *
     * @param collectionId the ID of the collection index to be retrieved.
     * @return the collection index with the given ID.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    CollectionIndex getCollectionIndexImpl(String collectionId) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNullOrEmpty(collectionId, "collectionId");

        CollectionIndex collection = null;

        // Obtain the database connection
        Connection conn = getConnection();

        try {
            try {
                // Retrieve all document source IDs in the collection index
                Set documentIds = getWordSourceIdsInCollection(collectionId, conn);

                // When the collection index exists in the persistence
                if (documentIds != null) {
                    Map wordMap = new HashMap();

                    // Retrieve all document indices in the set and add them to the word map
                    for (Iterator iter = documentIds.iterator(); iter.hasNext();) {
                        IndexPersistenceHelper
                            .addDocumentWordMap(wordMap, getDocumentIndex((WordSourceId) iter.next()));
                    }

                    // Create the collection index
                    collection = new CollectionIndex(null, wordMap, documentIds, collectionId);
                }
            } finally {
                conn.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Retrieving the collection index fails.", e);
        }

        return collection;
    }

    /**
     * Increases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *             not persisted.
     */
    void increaseDocumentUseCountImpl(WordSourceId wordSourceId) throws IndexPersistenceException {
        changeDocumentUseCount(wordSourceId, INCREASE_DOCUMENT_USE_COUNT);
    }

    /**
     * Decreases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *             not persisted; or the collection reference count is already 0.
     */
    void decreaseDocumentUseCountImpl(WordSourceId wordSourceId) throws IndexPersistenceException {
        changeDocumentUseCount(wordSourceId, DECREASE_DOCUMENT_USE_COUNT);
    }

    /**
     * Changes the collection reference count of the document index by executing the given SQL statement. When the
     * database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is changed.
     * @param sql the SQL statement used to change the reference count.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if changing the collection reference count fails; or the document index is not
     *             persisted; or the collection reference count is already 0 when trying to decrease the count.
     */
    private void changeDocumentUseCount(WordSourceId wordSourceId, String sql) throws IndexPersistenceException {
        String id = IndexPersistenceHelper.generateId(wordSourceId);

        // Obtain the database connection and start the transaction
        Connection conn = getConnection();

        beginTransaction(conn);

        try {
            // Execute the SQL statement to change the document use count
            PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, sql, new String[] {id});
            boolean success;

            try {
                success = (statement.getUpdateCount() != 0);
            } finally {
                statement.close();
            }

            if (!success) {
                // When no record is updated, either the document index is not persisted
                // or try to decrease a use count which is already 0
                throw new SQLException("The document index is not persisted or corrupted; "
                    + "or the document index use count cannot be changed to negative.");
            }

        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Error occurs when updating the collection index.", e);
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     * @return the reference count of the document index.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *             not persisted.
     */
    int getDocumentUseCountImpl(WordSourceId wordSourceId) throws IndexPersistenceException {
        String id = IndexPersistenceHelper.generateId(wordSourceId);

        // Obtain the database connection and start the transaction
        Connection conn = getConnection();
        int useCount;

        try {
            try {
                // Select the document use count
                PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_DOCUMENT_USE_COUNT,
                    new String[] {id});

                try {
                    // Get the use count for the document
                    ResultSet rs = statement.getResultSet();

                    if (!rs.next()) {
                        // When no record is selected
                        throw new SQLException("The document index is not persisted.");
                    }

                    useCount = rs.getInt(1);
                } finally {
                    // Close the statement even if error occurs
                    statement.close();
                }
            } finally {
                // Close the connection even if error occurs
                conn.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error occurs when updating the collection index.", e);
        }

        if (useCount < 0) {
            // Negative use count is not allowed
            throw new IndexPersistenceException("The use count cannot be negative.");
        }

        return useCount;
    }

    /**
     * Gets a set of all IDs of documents whose indices are persisted in the persistence.
     *
     * @return a set of all IDs of persisted document indices.
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    Set getIndexedDocumentsImpl() throws IndexPersistenceException {
        Set set;

        // Obtain the database connection
        Connection conn = getConnection();

        try {
            try {
                // Retrieve all persisted document index IDs
                set = getWordSourceIds(conn, SELECT_ALL_DOCUMENT, new Object[0]);
            } finally {
                // Close the connection even if error occurs
                conn.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Error occurs when retrieving the persisted document IDs.", e);
        }

        return set;
    }

    /**
     * Gets the document IDs from the persistence. The document IDs are obtained by executing the given SQL statement
     * and the parameters.
     *
     * @param conn the database connection used to retrieve the document IDs.
     * @param sql the SQL statement used to retrieve the document IDs.
     * @param params the parameters of the SQL statement used to retrieve the document IDs.
     * @return a set of document IDs retrieved from the persistence.
     * @throws SQLException if accessing the database fails.
     * @throws IndexPersistenceException if the retrieved document ID is not valid; or the document source cannot be
     *             retrieved.
     */
    private static Set getWordSourceIds(Connection conn, String sql, Object[] params) throws SQLException,
        IndexPersistenceException {
        Set set = new HashSet();

        // Execute the statement to retrieve necessary information
        PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, sql, params);

        try {
            ResultSet rs = statement.getResultSet();

            // Repeat until there is no more record.
            while (rs.next()) {
                // Parse the locale
                String[] localeStr = rs.getString(LOCALE_COLUMN).split(LOCALE_DELIMITER);
                String language = localeStr[0];
                String country = (localeStr.length > 1) ? localeStr[1] : EMPTY;
                String variant = (localeStr.length > 2) ? localeStr[2] : EMPTY;

                Object source;

                try {
                    InputStream is = rs.getBlob(SOURCE_COLUMN).getBinaryStream();
                    ObjectInputStream ois = null;

                    try {
                        // Parse the document source
                        ois = new ObjectInputStream(is);

                        source = ois.readObject();
                        ois.close();
                    } finally {
                        // Close the input stream
                        try {
                            if (ois == null) {
                                // When the input stream is created, but the object input stream fails
                                is.close();
                            } else {
                                ois.close();
                            }
                        } catch (NullPointerException e) {
                            // Informix JDBC Driver 3.00 JC3 workaround.
                            // When close the input stream of a BLOB, it sometime throws NullPointerException.
                        }
                    }
                } catch (IOException e) {
                    throw new IndexPersistenceException("The document source cannot be retrieved.", e);
                } catch (ClassNotFoundException e) {
                    throw new IndexPersistenceException("The document source cannot be deserialized.", e);
                }

                // Retrieve the delimiters
                PreparedStatement delimiterStatement = DBPersistenceHelper.executeSQL(conn, SELECT_DOCUMENT_DELIMITER,
                    new String[] {rs.getString(DOCUMENT_ID_COLUMN)});
                List delimiters = new ArrayList();

                try {
                    // For each delimiter, add it to the list
                    ResultSet delimiterSet = delimiterStatement.getResultSet();

                    while (delimiterSet.next()) {
                        delimiters.add(delimiterSet.getString(1));
                    }
                } finally {
                    delimiterStatement.close();
                }

                try {
                    // Add the newly created document source ID to the set
                    set.add(new WordSourceId(source, rs.getString(CLASS_NAME_COLUMN), (String[]) delimiters
                        .toArray(new String[0]), new Locale(language, country, variant)));
                } catch (IllegalArgumentException e) {
                    throw new IndexPersistenceException("The persisted document ID is not valid.", e);
                }
            }
        } finally {
            statement.close();
        }

        return set;
    }

    /**
     * Gets the document IDs in a collection index. The ID of the collection index is given. When the collection index
     * does not exist in the persistence, <code>null</code> is returned.
     *
     * @param collectionId the ID of the collection index whose document IDs are retrieved.
     * @param conn the database connection used to retrieve the document IDs.
     * @return a set of document IDs belonging to the collection index.
     * @throws SQLException if accessing the database fails.
     * @throws IndexPersistenceException if the retrieved document ID is not valid; or the document source cannot be
     *             retrieved.
     */
    static Set getWordSourceIdsInCollection(String collectionId, Connection conn) throws SQLException,
        IndexPersistenceException {
        return existCollection(collectionId, conn) ? getWordSourceIds(conn, SELECT_DOCUMENT_INDEX_COLLECTION,
            new String[] {collectionId}) : null;
    }

    /**
     * Adds the document source ID into the persistence.
     *
     * @param documentId the string ID of the document.
     * @param sourceId the source ID of the document.
     * @param conn the connection used to persist the document source ID.
     * @throws SQLException if persisting the document source ID fails.
     * @throws IndexPersistenceException if the source instance of the document is not serializable.
     */
    static void addWordSourceId(String documentId, WordSourceId sourceId, Connection conn) throws SQLException,
        IndexPersistenceException {
        // Persist the document ID without the delimiters and the document source
        DBPersistenceHelper
            .executeSQL(
                conn,
                ADD_DOCUMENT_INDEX_DOCUMENT,
                new Object[] {documentId, sourceId.getSourceLocale().toString(), sourceId.getSourceClassName(),
                    new byte[0]}).close();

        // Retrieve and update the document source, since there is no way to create a new BLOB instance
        PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_DOCUMENT_INDEX_SOURCE,
            new String[] {documentId});

        try {
            // Get the BLOB column for update
            ResultSet rs = statement.getResultSet();
            Blob data;

            rs.next();
            data = rs.getBlob(1);

            try {
                // Serialize the document source
                OutputStream os;

                try {
                    // This is a bug of Informix JDBC Driver 3.00 JC3
                    os = data.setBinaryStream(0);
                } catch (SQLException e) {
                    // This is the correct way of JDBC specification.
                    os = data.setBinaryStream(1);
                }

                ObjectOutputStream oos = null;

                try {
                    oos = new ObjectOutputStream(os);
                    oos.writeObject(sourceId.getSourceIdentity());
                } finally {
                    // Close the output stream even if error occurs.
                    if (oos == null) {
                        os.close();
                    } else {
                        oos.close();
                    }
                }
            } catch (IOException e) {
                throw new IndexPersistenceException("The document source cannot be serialized.", e);
            }

            // Update the BLOB column
            DBPersistenceHelper.executeSQL(conn, UPDATE_DOCUMENT_INDEX_DOCUMENT, new Object[] {data, documentId})
                .close();
        } finally {
            statement.close();
        }

        // Persist the delimiters
        String[] delimiters = sourceId.getDelimiters();

        for (int delimiterIndex = 0; delimiterIndex < delimiters.length; ++delimiterIndex) {
            DBPersistenceHelper.executeSQL(conn, ADD_DOCUMENT_INDEX_DELIMITER,
                new Object[] {documentId, delimiters[delimiterIndex], new Integer(delimiterIndex)}).close();
        }
    }
}

