/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * Defines utilities used in this package.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
final class DBPersistenceHelper {
    /**
     * Creates an instance of <code>DBPersistenceHelper</code> class. This private constructor prevents the creation
     * of a new instance.
     */
    private DBPersistenceHelper() {
    }

    /**
     * Validates the value of a variable with type <code>Object</code>. The value cannot be <code>null</code>.
     *
     * @param value value of the variable.
     * @param name name of the variable.
     * @throws IllegalArgumentException if <code>value</code> is <code>null</code>.
     */
    public static void validateNotNull(Object value, String name) {
        if (value == null) {
            throw new IllegalArgumentException(name + " cannot be null.");
        }
    }

    /**
     * Validates the value of a string variable. The value cannot be <code>null</code> and empty string.
     *
     * @param value value of the variable.
     * @param name name of the variable.
     * @throws IllegalArgumentException if <code>value</code> is <code>null</code> or empty string.
     */
    public static void validateNotNullOrEmpty(String value, String name) {
        validateNotNull(value, name);

        if (value.trim().length() == 0) {
            throw new IllegalArgumentException(name + " cannot be empty string.");
        }
    }

    /**
     * Executes the given SQL statement using the database connection and the given array of parameters. In case of a
     * SQL error, the SQL statement is closed.
     *
     * @param conn the database connection used to execute the SQL statement.
     * @param sql the SQL statement to be executed.
     * @param params the parameters of the SQL statement.
     * @return the SQL statement after execution.
     * @throws SQLException if executing the SQL statement fails.
     */
    public static PreparedStatement executeSQL(Connection conn, String sql, Object[] params) throws SQLException {
        PreparedStatement statement = conn.prepareStatement(sql);

        try {
            // Set parameters
            for (int paramIndex = 0; paramIndex < params.length; ++paramIndex) {
                Object obj = params[paramIndex];

                if (obj instanceof String) {
                    // This is a bug of Informix JDBC Driver 3.00 JC3.
                    statement.setString(paramIndex + 1, (String) obj);
                } else {
                    statement.setObject(paramIndex + 1, params[paramIndex]);
                }
            }

            // Execute the statement
            statement.execute();
        } catch (SQLException e) {
            // In case there is some error, close the SQL statement.
            statement.close();

            throw e;
        }

        return statement;
    }
}











