/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.CollationKey;
import java.text.Collator;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.persistence.impl.PersistenceConfigurationException;
import com.topcoder.document.index.wordsource.WordSourceId;

/**
 * Implements <code>IndexPersistence</code> interface by extending <code>AbstractDBIndexPersistence</code> class. It
 * provides the functionality to persist document indices and collection indices into databases. The source identity
 * instance of the document index must be serializable. In this implementation, the word and position data is stored in
 * two separate tables. Though it is slower than <code>FastAccessDBIndexPersistence</code>, it is more flexible, since
 * information can be retrieved directly.<br>
 * The DDL of the tables used by this class is shown as follows:<br>
 *
 * <pre>
 *  CREATE TABLE DOCUMENT_COLLECTION_DOCUMENT_XREF
 *  (
 *  COLLECTION_ID VARCHAR(40),
 *  DOCUMENT_ID VARCHAR(40),
 *  FOREIGN KEY (COLLECTION_ID) REFERENCES DOCUMENT_COLLECTION,
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT
 *  )
 *
 *  CREATE TABLE DOCUMENT_COLLECTION
 *  (
 *  COLLECTION_ID VARCHAR(40) PRIMARY KEY
 *  )
 *
 *  CREATE TABLE DOCUMENT
 *  (
 *  DOCUMENT_ID VARCHAR(40) PRIMARY KEY,
 *  LOCALE VARCHAR(30),
 *  SOURCE_IDENTITY BLOB,
 *  CLASS_NAME VARCHAR(255),
 *  USE_COUNT INTEGER
 *  )
 *
 *  CREATE TABLE DELIMITER
 *  (
 *  DOCUMENT_ID VARCHAR(40),
 *  DELIMITER_ENTRY VARCHAR(10),
 *  DELIMITER_SORT_ORDER INTEGER,
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT
 *  )
 *
 *  CREATE TABLE WORD
 *  (
 *  WORD_ID INTEGER,
 *  DOCUMENT_ID VARCHAR(40),
 *  WORD_ENTRY VARCHAR(255),
 *  FOREIGN KEY (DOCUMENT_ID) REFERENCES DOCUMENT.
 *  PRIMARY KEY (WORD_ID, DOCUMENT_ID)
 *  )
 *
 *  CREATE TABLE WORD_POSITION
 *  (
 *  DOCUMENT_ID VARCHAR(40),
 *  WORD_ID INTEGER,
 *  WORD_POSITION INTEGER,
 *  FOREIGN KEY (WORD_ID, DOCUMENT_ID) REFERENCES WORD
 *  )
 * </pre>
 *
 * This class is configurable. The configuration can be seen in <code>AbstractDBIndexPersistence</code> class.<br>
 * This class is stateless. However, its thread-safety depends on its super-class.
 *
 * @author AleaActaEst, TCSDEVELOPER
 * @version 1.0
 * @see AbstractDBIndexPersistence
 */
public class SimpleAccessDBIndexPersistence extends AbstractDBIndexPersistence {
    /**
     * Represents the SQL statement used to add a word entry into the persistence.
     */
    private static final String ADD_WORD_ENTRY = "INSERT INTO WORD (WORD_ID, DOCUMENT_ID, WORD_ENTRY) VALUES (?, ?, ?)";

    /**
     * Represents the SQL statement used to add a word position into the persistence.
     */
    private static final String ADD_WORD_POSITION = "INSERT INTO WORD_POSITION (DOCUMENT_ID, WORD_ID, WORD_POSITION)"
        + " VALUES (?, ?, ?)";

    /**
     * Represents the SQL statement used to select the words and word positions of a document index from the
     * persistence.
     */
    private static final String SELECT_WORDS = "SELECT WORD.WORD_ENTRY, WORD_POSITION.WORD_POSITION, WORD.WORD_ID"
        + " FROM (WORD LEFT OUTER JOIN WORD_POSITION ON WORD.WORD_ID = WORD_POSITION.WORD_ID"
        + " AND WORD.DOCUMENT_ID = WORD_POSITION.DOCUMENT_ID) RIGHT OUTER JOIN"
        + " DOCUMENT ON DOCUMENT.DOCUMENT_ID = WORD.DOCUMENT_ID WHERE DOCUMENT.DOCUMENT_ID = ?";

    /**
     * Represents the index of WORD_ENTRY column.
     */
    private static final int WORD_ENTRY_INDEX = 1;

    /**
     * Represents the index of WORD_POSITION column.
     */
    private static final int WORD_POSITION_INDEX = 2;

    /**
     * Represents the index of WORD_ID column.
     */
    private static final int WORD_ID_INDEX = 3;

    /**
     * Represents the array of table names whose records should be deleted when a document index is removed.
     */
    private static final String[] REMOVE_TABLES = new String[] {"WORD_POSITION", "WORD", "DELIMITER", "DOCUMENT"};

    /**
     * Creates a new instance of <code>SimpleAccessDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are loaded from the given namespace in the configuration.
     *
     * @param namespace the namespace where the configurations are loaded.
     * @throws IllegalArgumentException if <code>namespace</code> is <code>null</code> or empty string.
     * @throws PersistenceConfigurationException if the configurations are not valid; or accessing the configuration
     *             manager fails.
     */
    public SimpleAccessDBIndexPersistence(String namespace) throws PersistenceConfigurationException {
        super(namespace);
    }

    /**
     * Creates a new instance of <code>SimpleAccessDBIndexPersistence</code> class. The DB connection factory and the
     * connection name are given. When the connection name is empty string or <code>null</code>, the default
     * connection name is used.
     *
     * @param connectionFactory the DB connection factory used to create database connections.
     * @param connectionName the connection name used to create database connections.
     * @throws IllegalArgumentException if <code>connectionFactory</code> is <code>null</code>.
     */
    public SimpleAccessDBIndexPersistence(DBConnectionFactory connectionFactory, String connectionName) {
        super(connectionFactory, connectionName);
    }

    /**
     * Adds the given document index to the persistence. The initial collection reference count should be 0. When the
     * document index exists in the persistence, it fails. When the database operation fails, the modifications are
     * rolled back.
     *
     * @param documentIndex the document index to be persisted.
     * @throws IllegalArgumentException if <code>documentIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the document index fails; or the document index exists in the
     *             persistence; or the source identity instance is not serializable.
     */
    public void addDocumentIndex(DocumentIndex documentIndex) throws IndexPersistenceException {
        DBPersistenceHelper.validateNotNull(documentIndex, "documentIndex");

        // Get the ID and the data
        WordSourceId sourceId = documentIndex.getWordSourceId();
        String id = IndexPersistenceHelper.generateId(sourceId);

        // Obtain the database connection
        Connection conn = getConnection();

        // Begin the transaction
        beginTransaction(conn);

        try {
            // Add the document source ID to the table DOCUMENT
            addWordSourceId(id, sourceId, conn);

            // Add all words to the table WORD
            Map wordMap = documentIndex.getWords();
            int wordId = 1;

            for (Iterator iter = wordMap.entrySet().iterator(); iter.hasNext(); ++wordId) {
                Entry entry = (Entry) iter.next();

                // Add one word using the current word ID
                DBPersistenceHelper.executeSQL(conn, ADD_WORD_ENTRY,
                    new String[] {Integer.toString(wordId), id, ((CollationKey) entry.getKey()).getSourceString()})
                    .close();

                // Add all word positions to the table WORD_POSITION
                for (Iterator posIter = ((List) entry.getValue()).iterator(); posIter.hasNext();) {
                    DBPersistenceHelper.executeSQL(conn, ADD_WORD_POSITION,
                        new Object[] {id, Integer.toString(wordId), posIter.next()}).close();
                }
            }
        } catch (SQLException e) {
            // When error occurs, roll back the transaction
            rollbackTransaction(conn);

            throw new IndexPersistenceException("Adding the document index fails.", e);
        }

        // When all passes, commit the transaction
        commitTransaction(conn);
    }

    /**
     * Gets the document index from the persistence. The ID of the document is given. When the document is not persisted
     * in the persistence, <code>null</code> should be returned.
     *
     * @param wordSourceId the ID of the document whose index is retrieved.
     * @return the document index of the document with the ID.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the document index fails; or the source identity instance is not
     *             serializable.
     */
    public DocumentIndex getDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        String id = IndexPersistenceHelper.generateId(wordSourceId);
        DocumentIndex index = null;

        // Get the comparator
        Collator collator = Collator.getInstance(wordSourceId.getSourceLocale());

        collator.setStrength(Collator.PRIMARY);

        // Obtain the database connection
        Connection conn = getConnection();

        try {
            try {
                // Retrieve the words and positions of the document index
                PreparedStatement statement = DBPersistenceHelper.executeSQL(conn, SELECT_WORDS, new String[] {id});

                try {
                    ResultSet rs = statement.getResultSet();

                    if (rs.next()) {
                        // When there is a record in the result set, the document index exists
                        Map wordMap = new HashMap();
                        Map wordIdMap = new HashMap();

                        do {
                            // Get the word and its ID
                            String word = rs.getString(WORD_ENTRY_INDEX);
                            int wordId = rs.getInt(WORD_ID_INDEX);

                            if (!rs.wasNull()) {
                                // All null word should be ignored, since they are produced by outer join
                                CollationKey key = collator.getCollationKey(word);
                                List positions = (List) wordMap.get(key);
                                Integer currentWordId = (Integer) wordIdMap.get(key);

                                // Check duplicate words
                                if (currentWordId == null) {
                                    // The word never appears, remember its ID
                                    wordIdMap.put(key, new Integer(wordId));
                                } else if (currentWordId.intValue() != wordId) {
                                    // The word appears before, but has a different ID, duplicate words found
                                    throw new IndexPersistenceException("A word appears twice in the persistence.");
                                }

                                if (positions == null) {
                                    // The word does not exist, create a new list
                                    positions = new ArrayList();

                                    wordMap.put(key, positions);
                                }

                                int position = rs.getInt(WORD_POSITION_INDEX);

                                if (rs.wasNull()) {
                                    // The word has no position, throw error
                                    throw new IndexPersistenceException("A word must have a position.");
                                }

                                // Mark the word's position in the document
                                positions.add(new Integer(position));
                            }

                            // Process the next
                        } while (rs.next());

                        try {
                            index = new DocumentIndex(wordSourceId, wordMap);
                        } catch (IllegalArgumentException e) {
                            throw new IndexPersistenceException("The persisted document index is illegal.", e);
                        }
                    }
                } finally {
                    statement.close();
                }
            } finally {
                conn.close();
            }
        } catch (SQLException e) {
            throw new IndexPersistenceException("Retrieving the document index fails.", e);
        }

        return index;
    }

    /**
     * Removes a document index from the persistence. The document index can be deleted only when there is no more
     * collection containing the document index. The ID of the document is given. When the document is not persisted, it
     * fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose index is deleted.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if deleting the document index fails; or the document index is not persisted;
     *             or there is at least one collection containing the document index; or the source identity instance is
     *             not serializable.
     */
    public void removeDocumentIndex(WordSourceId wordSourceId) throws IndexPersistenceException {
        removeDocumentIndexImpl(wordSourceId, REMOVE_TABLES);
    }

    /**
     * Adds the collection index to the persistence. When the collection index exists in the persistence, it fails. When
     * the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be persisted.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if persisting the collection index fails; or the collection index exists in the
     *             persistence; or the source identity instance of one document index is not serializable.
     */
    public void addCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        addCollectionIndexImpl(collectionIndex);
    }

    /**
     * Gets the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, <code>null</code> should be returned.
     *
     * @param collectionId the ID of the collection index to be retrieved.
     * @return the collection index with the given ID.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if retrieving the collection index fails.
     */
    public CollectionIndex getCollectionIndex(String collectionId) throws IndexPersistenceException {
        return getCollectionIndexImpl(collectionId);
    }

    /**
     * Removes the collection index from the persistence. The ID of the collection index is given. When the collection
     * index is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionId the ID of the collection index to be deleted.
     * @throws IllegalArgumentException if <code>collectionId</code> is <code>null</code> or empty string.
     * @throws IndexPersistenceException if deleting the collection index fails; or the collection index ID cannot be
     *             found in the persistence.
     */
    public void removeCollectionIndex(String collectionId) throws IndexPersistenceException {
        removeCollectionIndexImpl(collectionId);
    }

    /**
     * Updates the collection index in the persistence. When the collection index ID cannot be found in the persistence,
     * it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param collectionIndex the collection index to be updated.
     * @throws IllegalArgumentException if <code>collectionIndex</code> is <code>null</code>.
     * @throws IndexPersistenceException if updating the collection index fails; or the collection index ID cannot be
     *             found in the persistence; or the source identity instance of one document index is not serializable.
     */
    public void updateCollectionIndex(CollectionIndex collectionIndex) throws IndexPersistenceException {
        updateCollectionIndexImpl(collectionIndex);
    }

    /**
     * Increases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is increased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if increasing the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public void increaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        increaseDocumentUseCountImpl(wordSourceId);
    }

    /**
     * Decreases the collection reference count of the document index. The document ID is given. When the document index
     * is not persisted, it fails. When the database operation fails, the modifications are rolled back.
     *
     * @param wordSourceId the ID of the document whose collection reference count is decreased by 1.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if decreasing the collection reference count fails; or the document index is
     *             not persisted; or the collection reference count is already 0; or the source identity instance is not
     *             serializable.
     */
    public void decreaseDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        decreaseDocumentUseCountImpl(wordSourceId);
    }

    /**
     * Gets a set of all IDs of documents whose indices are persisted in the persistence.
     *
     * @return a set of all IDs of persisted document indices.
     * @throws IndexPersistenceException if retrieving the document IDs fails.
     */
    public Set getIndexedDocuments() throws IndexPersistenceException {
        return getIndexedDocumentsImpl();
    }

    /**
     * Gets the collection reference count of the given document index. When the document index is not persisted, it
     * fails.
     *
     * @param wordSourceId the ID of the document index whose collection reference count is retrieved.
     * @return the reference count of the document index.
     * @throws IllegalArgumentException if <code>wordSourceId</code> is <code>null</code>.
     * @throws IndexPersistenceException if retrieving the collection reference count fails; or the document index is
     *             not persisted; or the source identity instance is not serializable.
     */
    public int getDocumentUseCount(WordSourceId wordSourceId) throws IndexPersistenceException {
        return getDocumentUseCountImpl(wordSourceId);
    }
}











