/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import java.text.Collator;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>IndexPersistenceHelper</code> class. This class is a helper class. Its
 * all static methods are tested.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class IndexPersistenceHelperTestCase extends TestCase {
    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(IndexPersistenceHelperTestCase.class);
    }

    /**
     * Tests <code>generateId(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGenerateIdWordSourceIdNull() throws Exception {
        try {
            IndexPersistenceHelper.generateId(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>generateId(WordSourceId)</code> when the source identity instance is not serializable. Expect
     * IndexPersistenceException.
     */
    public void testGenerateIdWordSourceIdNotSerializable() {
        WordSourceId id = new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US);

        try {
            IndexPersistenceHelper.generateId(id);
            fail("The source identity instance is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>generateId(WordSourceId)</code> when the document IDs are the same. The generated
     * string IDs should be the same. However, when the document IDs are not the same, it still has minor chance to have
     * the same IDs.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGenerateIdWordSourceIdAccuracy() throws Exception {
        WordSourceId id1 = new WordSourceId("word", "class", new String[] {" "}, Locale.US);
        WordSourceId id2 = new WordSourceId("word", "class", new String[] {" "}, Locale.US);

        // Verify
        assertEquals("The string ID of the same document ID should be the same.", IndexPersistenceHelper
            .generateId(id1), IndexPersistenceHelper.generateId(id2));
    }

    /**
     * Tests accuracy of <code>addDocumentWordMap(Map, DocumentIndex)</code>. The words in the document index is
     * accumulated into the collection word map. This is verified by comparing the result to an expected map.
     */
    public void testAddDocumentWordMapMapDocumentIndexAccuracy() {
        // Build the existing collection word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);
        WordSourceId id1 = new WordSourceId("word1", "class", new String[] {" "}, Locale.US);
        Set set = new HashSet();

        collator.setStrength(Collator.PRIMARY);
        set.add(id1);

        wordMap.put(collator.getCollationKey("word1"), set);

        // Build the document index
        WordSourceId id2 = new WordSourceId("word2", "class", new String[] {" "}, Locale.US);
        Map map = new HashMap();

        map.put(collator.getCollationKey("WorD1"), Collections.singletonList(new Integer(0)));
        map.put(collator.getCollationKey("WorD2"), Collections.singletonList(new Integer(1)));

        DocumentIndex index = new DocumentIndex(id2, map);

        // Build the expected collection word map
        Map expected = new HashMap(wordMap);

        ((Set) expected.get(collator.getCollationKey("word1"))).add(id2);
        expected.put(collator.getCollationKey("word2"), Collections.singleton(id2));

        // Add the document index to the collection word map
        IndexPersistenceHelper.addDocumentWordMap(wordMap, index);

        // Verify
        assertEquals("The document index word map should be added to the collection word map.", expected, wordMap);
    }
}








