/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.lang.reflect.Field;
import java.nio.channels.FileChannel;

/**
 * Defines utilities used in test cases.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public final class TestHelper {
    /**
     * Creates a new instance of <code>TestHelper</code> class. This private constructor prevents the creation of a
     * new instance.
     */
    private TestHelper() {
    }

    /**
     * Gets the field value of the given instance. The field name is also given.
     *
     * @param instance the instance whose field value is retrieved.
     * @param field the field name.
     * @return the field value of the given instance.
     * @throws Exception if error occurs when retrieving the value.
     */
    public static Object getField(Object instance, String field) throws Exception {
        return getField(instance.getClass(), instance, field);
    }

    /**
     * Gets the field value of the given instance. The field name and the class containing the field is also given.
     *
     * @param type the class containing the field to be retrieved.
     * @param instance the instance whose field value is retrieved.
     * @param field the field name.
     * @return the field value of the given instance.
     * @throws Exception if error occurs when retrieving the value.
     */
    public static Object getField(Class type, Object instance, String field) throws Exception {
        Field info = type.getDeclaredField(field);
        Object value;

        try {
            info.setAccessible(true);
            value = info.get(instance);
        } finally {
            info.setAccessible(false);
        }

        return value;
    }

    /**
     * Copies the content of one file to another. The destination file is overwritten.
     *
     * @param source the source file to be copied.
     * @param dest the destination file to be copied to.
     * @throws IOException if I/O error occurs when copying the file.
     */
    public static void copy(File source, File dest) throws IOException {
        FileChannel input = new FileInputStream(source).getChannel();

        try {
            FileChannel output = new FileOutputStream(dest).getChannel();

            try {
                // Copy the file using channel.
                input.transferTo(0, input.size(), output);
            } finally {
                output.close();
            }
        } finally {
            input.close();
        }
    }

    /**
     * Serializes an instance into an array of bytes.
     *
     * @param obj the instance to be serialized.
     * @return the array of bytes representing the serialized instance.
     * @throws IOException if the object cannot be serialized.
     */
    public static byte[] serialize(Object obj) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(baos);

        oos.writeObject(obj);
        oos.close();

        return baos.toByteArray();
    }
}








