/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.db;

import java.sql.Blob;
import java.sql.Clob;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import com.topcoder.db.connectionfactory.DBConnectionFactory;
import com.topcoder.db.connectionfactory.DBConnectionFactoryImpl;
import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;
import com.topcoder.util.config.ConfigManager;

import junit.framework.Test;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>FastAccessDBIndexPersistence</code> class.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class FastAccessDBIndexPersistenceTestCase extends CommonDBIndexPersistenceTestCase {
    /**
     * Represents the configuration manager used in tests.
     */
    private static final ConfigManager CONFIG = ConfigManager.getInstance();

    /**
     * Represents the DB connection factory used in tests.
     */
    private DBConnectionFactory factory;

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(FastAccessDBIndexPersistenceTestCase.class);
    }

    /**
     * Creates a new instance of the DB index persistence used in tests.
     *
     * @return a new instance of the DB index persistence to be tested.
     * @throws Exception if error occurs when creating the index persistence.
     */
    protected AbstractDBIndexPersistence createPersistence() throws Exception {
        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        CONFIG.add("DBFactory.xml");
        factory = new DBConnectionFactoryImpl("com.topcoder.db.connectionfactory.DBConnectionFactoryImpl");

        for (Iterator iter = CONFIG.getAllNamespaces(); iter.hasNext();) {
            CONFIG.removeNamespace((String) iter.next());
        }

        return new FastAccessDBIndexPersistence(factory, null);
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the document index in the collection is illegal in the
     * persistence. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringIllegalDocumentIndex() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Add a collection index
        persistence
            .addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id"));

        // Update the document index, so that a word has no position
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word 1  word2  word3 2 3", stringId}).close();

        try {
            persistence.getCollectionIndex("id");
            fail("The document index file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>addDocumentIndex(DocumentIndex)</code>. A new document record should be created in
     * table DOCUMENT. Two delimiter records should be created in table DELIMITER. A word data record should be created
     * in table WORD_DATA. These are verified by SQL statement. The word data should be created in a proper format. This
     * is verified by comparing to a reference instance.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexAccuracy() throws Exception {
        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);
        Set wordData = new HashSet();

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);

            // Build the word data
            StringBuffer buffer = new StringBuffer("Word" + Integer.toString(i));

            for (Iterator iter = positions.iterator(); iter.hasNext();) {
                buffer.append(' ');
                buffer.append(iter.next());
            }

            wordData.add(buffer.toString());
        }

        // Create the document index
        WordSourceId sourceId = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(sourceId);
        DocumentIndex index = new DocumentIndex(sourceId, wordMap);

        // Persist it
        persistence.addDocumentIndex(index);

        // Verify the document record
        PreparedStatement statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT * FROM DOCUMENT WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            ResultSet rs = statement.getResultSet();

            rs.next();

            Blob blob = rs.getBlob("SOURCE_IDENTITY");

            assertEquals("The document locale should be persisted.", Locale.US.toString(), rs.getString("LOCALE"));
            assertEquals("The document class name should be persisted.", "class", rs.getString("CLASS_NAME"));
            assertEquals("The document use count should be 0.", 0, rs.getInt("USE_COUNT"));
            assertTrue("The document source identity should be persisted.", Arrays.equals(TestHelper.serialize("id"),
                blob.getBytes(1, (int) blob.length())));
        } finally {
            statement.close();
        }

        // Verify the delimiter
        List actual = new ArrayList();

        statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT DELIMITER_ENTRY FROM DELIMITER WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            // Read all delimiters belonging to this document index
            ResultSet rs = statement.getResultSet();

            while (rs.next()) {
                actual.add(rs.getString(1));
            }
        } finally {
            statement.close();
        }

        assertEquals("The delimiters should be persisted.", Arrays.asList(sourceId.getDelimiters()), actual);

        // Verify the word data
        String[] data;

        statement = DBPersistenceHelper.executeSQL(connection,
            "SELECT WORD_DATA FROM WORD_DATA WHERE DOCUMENT_ID = ?", new Object[] {stringId});

        try {
            ResultSet rs = statement.getResultSet();

            rs.next();

            Clob clob = rs.getClob(1);

            // Split the data using two consecutive delimiters
            data = clob.getSubString(1, (int) clob.length()).split("  ");
        } finally {
            statement.close();
        }

        // Each element in the data array should exist in the set
        assertEquals("The word data should be persisted in a proper format.", wordData,
            new HashSet(Arrays.asList(data)));
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the word data is not a number. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdPositionNotNumber() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a position is not a number
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word abc  word2 2 3", stringId}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is not a number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the word data is a negative number.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdPositionNegative() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a position is not a number
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word -1  word2 2 3", stringId}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is a negative number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the word position of a word is missing. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNoPosition() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a word has no position
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word  word2 2 3", stringId}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("The word position of a word is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word appears twice in the word data. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdWordDuplicate() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a word appears twice
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word 1  wORd 2 3", stringId}).close();

        try {
            persistence.getDocumentIndex(id);
            fail("A word appears twice in the word data, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests acuracy of <code>getDocumentIndex(WordSourceId)</code> when a word is empty string in the word data. It
     * should be read successfully, since empty word is allowed in the word map.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdWordEmptyAccuracy() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        // Add a document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Update the document index, so that a word is empty string
        DBPersistenceHelper.executeSQL(connection, "UPDATE WORD_DATA SET WORD_DATA = ? WHERE DOCUMENT_ID = ?",
            new Object[] {"word 1   4  word2 2 3", stringId}).close();

        // Get the document index
        DocumentIndex index = persistence.getDocumentIndex(id);

        assertTrue("The empty word should be read.", index.getWords().containsKey(collator.getCollationKey("")));
    }

    /**
     * Tests accuracy of <code>removeDocumentIndex(WordSourceId)</code>. All records related to this document index
     * in table DOCUMENT, DELIMITER, and WORD_DATA should be deleted.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\1"}, Locale.US);

        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Create the document index
        DocumentIndex index = new DocumentIndex(id, wordMap);

        // Persist the document index with multiple files
        persistence.addDocumentIndex(index);

        // Persist another document index with no word
        persistence.addDocumentIndex(new DocumentIndex(
            new WordSourceId("id1", "class", new String[] {" "}, Locale.PRC), Collections.EMPTY_MAP));

        // Remove the document index
        persistence.removeDocumentIndex(id);

        // Verify records in tables
        assertNoRecord("The index record in DOCUMENT should be deleted.",
            "SELECT * FROM DOCUMENT WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
        assertNoRecord("The index records in DELIMITER should be deleted.",
            "SELECT * FROM DELIMITER WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
        assertNoRecord("The index record in WORD_DATA should be deleted.",
            "SELECT * FROM WORD_DATA WHERE DOCUMENT_ID = ?", new Object[] {IndexPersistenceHelper.generateId(id)});
    }
}







