/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.document.index.persistence.impl.xml;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import com.topcoder.document.index.CollectionIndex;
import com.topcoder.document.index.DocumentIndex;
import com.topcoder.document.index.persistence.IndexPersistence;
import com.topcoder.document.index.persistence.IndexPersistenceException;
import com.topcoder.document.index.persistence.TestHelper;
import com.topcoder.document.index.persistence.impl.IndexPersistenceHelper;
import com.topcoder.document.index.wordsource.WordSourceId;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Tests functionality and error cases of <code>XmlIndexPersistence</code> class methods. Some error cases cannot be
 * produced automatically by a program, such as closing file error. These error cases are not tested in the test cases.
 *
 * @author TCSDEVELOPER
 * @version 1.0
 */
public class XmlIndexPersistenceMethodTestCase extends TestCase {
    /**
     * Represents the persistence base directory used in tests.
     */
    private static final File BASE_DIR = new File("test_files/Xml");

    /**
     * Represents the <code>XmlIndexPersistence</code> instance used in tests.
     */
    private IndexPersistence persistence;

    /**
     * Aggragates all tests in this class.
     *
     * @return test suite aggragating all tests.
     */
    public static Test suite() {
        return new TestSuite(XmlIndexPersistenceMethodTestCase.class);
    }

    /**
     * Gets the DOM document of the document usage XML file.
     *
     * @return the DOM document of the document usage file.
     * @throws Exception if loading the document usage file fails; or the document usage file corrupts.
     */
    private Document getDocumentUsage() throws Exception {
        File file = new File(BASE_DIR, "document_usage.xml");

        return DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(file);
    }

    /**
     * Deletes all XML files in the persistence directory.
     */
    private static void deleteXmlFiles() {
        // Delete all files in the persistence directory.
        File[] files = BASE_DIR.listFiles();

        for (int i = 0; i < files.length; ++i) {
            files[i].delete();
        }
    }

    /**
     * Sets up the test environment. The persistence directory is cleaned. The test instance is created. A valid
     * document usage file is copied.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    protected void setUp() throws Exception {
        deleteXmlFiles();
        persistence = new XmlIndexPersistence(1, 1023, BASE_DIR.getAbsolutePath());
        TestHelper.copy(new File("test_files/XmlReference/document_usage_valid.xml"), new File(BASE_DIR,
            "document_usage.xml"));
    }

    /**
     * Cleans up the test environment. The test instance is disposed. The persistence directory is cleaned.
     */
    protected void tearDown() {
        persistence = null;
        deleteXmlFiles();
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document index is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexNull() throws Exception {
        try {
            persistence.addDocumentIndex(null);
            fail("The document index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testAddDocumentIndexDocumentIndexSourceNotSerializable() {
        try {
            persistence.addDocumentIndex(new DocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "},
                Locale.US), Collections.EMPTY_MAP));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document index is already persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexPersisted() throws Exception {
        DocumentIndex index = new DocumentIndex(new WordSourceId("id", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP);

        // Persist the document index
        persistence.addDocumentIndex(index);

        try {
            // Persist it again
            persistence.addDocumentIndex(index);
            fail("The document index is already persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document index cannot be persisted due to IO error.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexIOError() throws Exception {
        DocumentIndex index = new DocumentIndex(new WordSourceId("id", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP);

        // Create a directory with the same name.
        new File(BASE_DIR, "doc_index_" + IndexPersistenceHelper.generateId(index.getWordSourceId()) + "_1.xml")
            .mkdir();

        try {
            persistence.addDocumentIndex(index);
            fail("The document index cannot be persisted due to IO error, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addDocumentIndex(DocumentIndex)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexDocumentUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.addDocumentIndex(new DocumentIndex(new WordSourceId("id", "class", new String[] {" "},
                Locale.US), Collections.EMPTY_MAP));
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>addDocumentIndex(DocumentIndex)</code>. The document index should be persisted as
     * several XML files. The file is split according to the file size margin. The document usage entry is created and
     * persisted. The document usage count should be 0. The accuracy of the document index files is verified by
     * retrieving the document index from the persistence. The word map should be exactly the same. The accuracy of the
     * document usage count is verified by the DOM parser. The document index files should be two file.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddDocumentIndexDocumentIndexAccuracy() throws Exception {
        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Create the document index
        DocumentIndex index = new DocumentIndex(new WordSourceId("id", "class", new String[] {"<", "\0"}, Locale.US),
            wordMap);

        // Persist it
        persistence.addDocumentIndex(index);

        // Verify the number of files. There should be three files in the directory
        assertEquals("The document index files should be split into two files.", 3, BASE_DIR.list().length);

        // Retrieve it
        DocumentIndex actual = persistence.getDocumentIndex(index.getWordSourceId());

        // Verify the word map
        assertEquals("The word map should be exactly the same.", index.getWords(), actual.getWords());

        // Verify the document usage count
        Document doc = getDocumentUsage();
        Element element = (Element) doc.getDocumentElement().getFirstChild();

        assertEquals("The ID of the document usage entry should be filled.", IndexPersistenceHelper.generateId(index
            .getWordSourceId()), element.getAttribute("document-id"));
        assertEquals("The document usage count should be 0.", "0", element.getAttribute("usage-count"));
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNull() throws Exception {
        try {
            persistence.getCollectionIndex(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testGetDocumentIndexWordSourceIdSourceNotSerializable() {
        try {
            persistence.getDocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the document index file is malformed. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdXMLMalform() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy a malformed file
        TestHelper.copy(new File("test_files/XmlReference/doc_index_malformed.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("The document index file is malformed, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the document index file is not a
     * number. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdPositionNotNumber() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy an invalid file
        TestHelper.copy(new File("test_files/XmlReference/doc_index_posnotnumber.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is not a number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word position in the document index file is a negative
     * number. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdPositionNegative() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy an invalid file
        TestHelper.copy(new File("test_files/XmlReference/doc_index_posnegative.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("The word position is a negative number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when the word position of a word is missing. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNoPosition() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy an invalid file
        TestHelper.copy(new File("test_files/XmlReference/doc_index_nopos.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("The word position of a word is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word is not properly escaped as '&92;uXXXX'. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdWordError() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy an invalid file
        TestHelper.copy(new File("test_files/XmlReference/doc_index_worderror.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("A word is not properly escaped, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when some files belonging to the document index are missing.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdFileMissing() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Copy a files. There would be one file (file 2) missing.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_1.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_1.xml"));
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_2.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_3.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("Some files of the document index are missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentIndex(WordSourceId)</code> when a word appears twice in the XML files.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdWordDuplicate() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Copy a files. There would be some words which appear twice across files
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_1.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_1.xml"));
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_2.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_2.xml"));
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_2.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_3.xml"));

        try {
            persistence.getDocumentIndex(id);
            fail("A word appears twice in the XML files, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getDocumentIndex(WordSourceId)</code> when the document index is not persisted. It
     * should return <code>null</code>.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdNotPersistedAccuracy() throws Exception {
        assertNull("The document index is not persisted, should return null.", persistence
            .getDocumentIndex(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US)));
    }

    /**
     * Tests accuracy of <code>getDocumentIndex(WordSourceId)</code> when the multiple document index files exist. It
     * should read all consecutive sequence files starting from 1. The word map of the loaded document index should be
     * exactly the same as a reference word map.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentIndexWordSourceIdAccuracy() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);
        String stringId = IndexPersistenceHelper.generateId(id);

        // Create the expected word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Copy the persisted files
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_1.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_1.xml"));
        TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_2.xml"), new File(BASE_DIR,
            "doc_index_" + stringId + "_2.xml"));

        // Load the document index
        DocumentIndex actual = persistence.getDocumentIndex(id);

        // Verify
        assertEquals("The loaded word map should be the same as the reference one.", wordMap, actual.getWords());
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdNull() throws Exception {
        try {
            persistence.removeDocumentIndex(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testRemoveDocumentIndexWordSourceIdSourceNotSerializable() {
        try {
            persistence.removeDocumentIndex(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testRemoveDocumentIndexWordSourceIdNotPersisted() {
        try {
            persistence.removeDocumentIndex(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when the usage count of the document index is 0. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdUsageCountNotZero() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Persist the document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Increase the usage count
        persistence.increaseDocumentUseCount(id);

        try {
            persistence.removeDocumentIndex(id);
            fail("The usage count of the document index is not 0, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeDocumentIndex(WordSourceId)</code> when a document index file cannot be deleted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdDeleteError() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Persist the document index with no words
        persistence.addDocumentIndex(new DocumentIndex(id, Collections.EMPTY_MAP));

        // Open one file, so that it cannot be deleted.
        InputStream is = new FileInputStream("test_files/Xml/doc_index_" + IndexPersistenceHelper.generateId(id)
            + "_1.xml");

        try {
            persistence.removeDocumentIndex(id);
            fail("One document index file cannot be deleted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        } finally {
            is.close();
        }
    }

    /**
     * Tests accuracy of <code>removeDocumentIndex(WordSourceId)</code>. All document index files related to this
     * document index should be deleted. The document usage entry in the document usage file should be deleted. These
     * are verified by listing all files in the persistence directory and DOM of the document usage file.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveDocumentIndexWordSourceIdAccuracy() throws Exception {
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Create the word map
        Map wordMap = new HashMap();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        for (int i = 0; i < 6; ++i) {
            List positions = new ArrayList();

            for (int j = 0; j < 10; ++j) {
                positions.add(new Integer(j));
            }

            wordMap.put(collator.getCollationKey("Word" + Integer.toString(i)), positions);
        }

        // Add another word containing illegal XML character
        wordMap.put(collator.getCollationKey("\1"), Collections.singletonList(new Integer(0)));

        // Create the document index
        DocumentIndex index = new DocumentIndex(id, wordMap);

        // Persist the document index with multiple files
        persistence.addDocumentIndex(index);

        // Persist another document index with no word
        persistence.addDocumentIndex(new DocumentIndex(
            new WordSourceId("id1", "class", new String[] {" "}, Locale.PRC), Collections.EMPTY_MAP));

        // Remove the document index
        persistence.removeDocumentIndex(id);

        // Should leave only two files in the directory, which is 'document_usage.xml' and another document index
        assertEquals("All document index files related to this one should be deleted.", 2, BASE_DIR.list().length);

        // Verify document usage entry, should leave only one entry
        Document doc = getDocumentUsage();

        assertEquals("The document usage entry should be deleted.", 1, doc.getChildNodes().getLength());
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the collection index is <code>null</code>.
     * Expect IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexNull() throws Exception {
        try {
            persistence.addCollectionIndex(null);
            fail("The collection index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the collection index is already persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexExist() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        // Persist it
        persistence.addCollectionIndex(collection);

        try {
            persistence.addCollectionIndex(collection);
            fail("The collection index is already persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the document index in the collection is not
     * persisted. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexDocumentIndexNotExist() throws Exception {
        // Create a document ID
        WordSourceId id = new WordSourceId("id", "class", new String[] {" "}, Locale.US);

        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.addCollectionIndex(collection);
            fail("The document index in the collection is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when I/O error occurs. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexIOError() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        // Create a directory with the same name as the collection index file
        new File(BASE_DIR, "doc_collection_id.xml").mkdir();

        try {
            persistence.addCollectionIndex(collection);
            fail("The document index in the collection is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when a document source identity is not serializable.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexSourceNotSerializable() throws Exception {
        WordSourceId id = new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US);

        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.addCollectionIndex(collection);
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>addCollectionIndex(CollectionIndex)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexDocumentUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        try {
            persistence.addCollectionIndex(collection);
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>addCollectionIndex(CollectionIndex)</code>. The document string IDs should be
     * persisted in the collection index file. The document collection inclusion should be updated. This is verified by
     * DOM.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testAddCollectionIndexCollectionIndexAccuracy() throws Exception {
        Set stringIds = new HashSet();
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        // Build the string IDs and document IDs
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            sourceIds.add(id);
            stringIds.add(IndexPersistenceHelper.generateId(id));

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            // Persist the document index
            persistence.addDocumentIndex(new DocumentIndex(id, map));
        }

        // Persist another document index which is not in the collection index
        persistence.addDocumentIndex(new DocumentIndex(new WordSourceId("no", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP));

        // Persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, sourceIds, "id"));

        // Verify collection inclusion
        Document usage = getDocumentUsage();
        NodeList nodes = usage.getDocumentElement().getChildNodes();
        Set actual = new HashSet();

        for (int i = 0; i < nodes.getLength(); ++i) {
            Element element = (Element) nodes.item(i);

            // Extract all collection inclusion
            if (element.getFirstChild().getChildNodes().getLength() == 1) {
                actual.add(element.getAttribute("document-id"));
            }
        }

        assertEquals("The collection inclusion should be added.", stringIds, actual);

        // Verify collection index file
        Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(
            new File("test_files/Xml/doc_collection_id.xml"));

        nodes = doc.getDocumentElement().getFirstChild().getChildNodes();
        actual.clear();

        for (int i = 0; i < nodes.getLength(); ++i) {
            actual.add(nodes.item(i).getFirstChild().getNodeValue());
        }

        assertEquals("The collection index file should contain all string IDs.", stringIds, actual);
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the collection ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringNull() throws Exception {
        try {
            persistence.getCollectionIndex(null);
            fail("The collection ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the collection ID is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringEmpty() throws Exception {
        try {
            persistence.getCollectionIndex("  ");
            fail("The collection ID is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when the document index in the collection is illegal in the
     * persistence. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringIllegalDocumentIndex() throws Exception {
        // Copy only the collection index file without document index files
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        try {
            persistence.getCollectionIndex("id");
            fail("The document index file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when an attribute of a document index in the collection is
     * missing. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringAttributeMissing() throws Exception {
        // Copy the collection index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id_single.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_noattr.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getCollectionIndex("id");
            fail("The document index file has attribute missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when a document index's delimiter is not escaped properly. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringDelimiterError() throws Exception {
        // Copy the collection index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id_single.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_delimitererror.xml"), new File(BASE_DIR,
            "doc_index_" + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getCollectionIndex("id");
            fail("The delimiter is not escaped properly, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when a document source identity cannot be deserialized. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringDeserializeError() throws Exception {
        // Copy the collection index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id_single.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_deserializeerror.xml"), new File(BASE_DIR,
            "doc_index_" + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getCollectionIndex("id");
            fail("The document source identity cannot be deserialized, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getCollectionIndex(String)</code> when a document ID cannot be created using the data. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringIllegalID() throws Exception {
        // Copy the collection index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id_single.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_iderror.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getCollectionIndex("id");
            fail("The document ID cannot be created using the data, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getCollectionIndex(String)</code> when the collection index is not persisted. It should
     * return <code>null</code>.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringNotExistAccuracy() throws Exception {
        assertNull("The document index is not persisted, should return null.", persistence.getCollectionIndex("id"));
    }

    /**
     * Tests accuracy of <code>getCollectionIndex(String)</code> when the collection index is persisted. The
     * collection index should be retrieved from the persistence. The document IDs should be read back. The word map
     * should be built according to the word maps in the document indices. These are verified by comparing the
     * collection index to reference instances.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetCollectionIndexStringExistAccuracy() throws Exception {
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);
        Map expected = new HashMap();

        collator.setStrength(Collator.PRIMARY);

        // Build the document IDs, and the expected word map
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);
            String stringId = IndexPersistenceHelper.generateId(id);

            sourceIds.add(id);

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            DocumentIndex index = new DocumentIndex(id, map);

            // Copy the document index persistence file
            TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_1.xml"), new File(BASE_DIR,
                "doc_index_" + stringId + "_1.xml"));

            // Build the expected word map
            IndexPersistenceHelper.addDocumentWordMap(expected, index);
        }

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Get the collection index
        CollectionIndex collection = persistence.getCollectionIndex("id");

        // Verify the word map and the document IDs
        assertEquals("The word map should be the same as reference.", expected, collection.getWords());
        assertEquals("The document IDs should be the same as reference.", sourceIds, collection.getAllDocumentIds());
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringNull() throws Exception {
        try {
            persistence.removeCollectionIndex(null);
            fail("The collection ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection ID is empty string. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringEmpty() throws Exception {
        try {
            persistence.removeCollectionIndex("  ");
            fail("The collection ID is empty string, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection index does not exist. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringNotExist() throws Exception {
        try {
            persistence.removeCollectionIndex("id");
            fail("The collection index does not exist, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringDocumentUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        try {
            persistence.removeCollectionIndex("id");
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection inclusion in the document usage file does
     * not contain this collection index. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringDocumentUsageMismatch() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_mismatch.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        try {
            persistence.removeCollectionIndex("id");
            fail("The document usage file mismatches the collection index, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>removeCollectionIndex(String)</code> when the collection index file cannot be deleted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringDeleteError() throws Exception {
        // Copy the document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_collection.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Open the file so that it cannot be deleted.
        InputStream is = new FileInputStream("test_files/Xml/doc_collection_id.xml");

        try {
            persistence.removeCollectionIndex("id");
            fail("The collection index file cannot be deleted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        } finally {
            is.close();
        }
    }

    /**
     * Tests accuracy of <code>removeCollectionIndex(String)</code>. The collection index file should be deleted.
     * This is verified by file. The document usage file should be updated. This is verified by DOM.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testRemoveCollectionIndexStringAccuracy() throws Exception {
        // Copy the document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_collection.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        // Remove the collection index
        persistence.removeCollectionIndex("id");

        // Verify the file
        assertFalse("The collection file should be deleted.", new File(BASE_DIR, "doc_collection_id.xml").isFile());

        // Verify collection inclusion
        Document usage = getDocumentUsage();
        NodeList nodes = usage.getDocumentElement().getChildNodes();
        Set actual = new HashSet();

        for (int i = 0; i < nodes.getLength(); ++i) {
            Element element = (Element) nodes.item(i);

            // Extract all collection inclusion
            if (element.getFirstChild().getChildNodes().getLength() == 1) {
                actual.add(element.getAttribute("document-id"));
            }
        }

        assertTrue("The collection inclusion should be deleted.", actual.isEmpty());
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the collection index is <code>null</code>.
     * Expect IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexNull() throws Exception {
        try {
            persistence.updateCollectionIndex(null);
            fail("The collection index is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the collection index is not persisted. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexNotExist() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The collection index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the document index in the collection is not
     * persisted. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexDocumentIndexNotExist() throws Exception {
        // Create a document ID
        WordSourceId id = new WordSourceId("id", "class", new String[] {" "}, Locale.US);

        // Create and persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id"));

        // Create the same collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document index in the collection is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when a document source identity is not serializable.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdatecollectionIndexCollectionIndexSourceNotSerializable() throws Exception {
        WordSourceId id = new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US);

        // Create and persist the collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id"));

        // Create the same collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.singleton(id), "id");

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(CollectionIndex)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexDocumentUsageError() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        // Persist it
        persistence.addCollectionIndex(collection);

        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>updateCollectionIndex(String)</code> when the collection inclusion in the document usage file does
     * not contain this collection index. Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexStringDocumentUsageMismatch() throws Exception {
        // Create a collection index
        CollectionIndex collection = new CollectionIndex(null, Collections.EMPTY_MAP, Collections.EMPTY_SET, "id");

        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_mismatch.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        // Copy the collection index reference file
        TestHelper.copy(new File("test_files/XmlReference/doc_collection_id.xml"), new File(BASE_DIR,
            "doc_collection_id.xml"));

        try {
            persistence.updateCollectionIndex(collection);
            fail("The document usage file mismatches the collection index, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>updateCollectionIndex(CollectionIndex)</code>. The document string IDs should be
     * persisted in the collection index file. The document collection inclusion should be updated. This is verified by
     * DOM.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testUpdateCollectionIndexCollectionIndexAccuracy() throws Exception {
        Set stringIds = new HashSet();
        Set sourceIds = new HashSet();
        Collator collator = Collator.getInstance(Locale.US);

        collator.setStrength(Collator.PRIMARY);

        // Build the string IDs and document IDs to be updated
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);

            sourceIds.add(id);
            stringIds.add(IndexPersistenceHelper.generateId(id));

            // Create a word map
            Map map = new HashMap();

            for (int j = 0; j < 4; ++j) {
                map.put(collator.getCollationKey("word" + (i / 2 + j)), Collections.singletonList(new Integer(1)));
            }

            // Persist the document index
            persistence.addDocumentIndex(new DocumentIndex(id, map));
        }

        // Persist another document index which is initially in the collection index
        WordSourceId singleId = new WordSourceId("no", "class", new String[] {" "}, Locale.US);

        persistence.addDocumentIndex(new DocumentIndex(new WordSourceId("no", "class", new String[] {" "}, Locale.US),
            Collections.EMPTY_MAP));

        // Persist the initial collection index
        persistence.addCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP,
            Collections.singleton(singleId), "id"));

        // Update the collection index
        persistence.updateCollectionIndex(new CollectionIndex(null, Collections.EMPTY_MAP, sourceIds, "id"));

        // Verify collection inclusion
        Document usage = getDocumentUsage();
        NodeList nodes = usage.getDocumentElement().getChildNodes();
        Set actual = new HashSet();

        for (int i = 0; i < nodes.getLength(); ++i) {
            Element element = (Element) nodes.item(i);

            // Extract all collection inclusion
            if (element.getFirstChild().getChildNodes().getLength() == 1) {
                actual.add(element.getAttribute("document-id"));
            }
        }

        assertEquals("The collection inclusion should be updated.", stringIds, actual);

        // Verify collection index file
        Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(
            new File("test_files/Xml/doc_collection_id.xml"));

        nodes = doc.getDocumentElement().getFirstChild().getChildNodes();
        actual.clear();

        for (int i = 0; i < nodes.getLength(); ++i) {
            actual.add(nodes.item(i).getFirstChild().getNodeValue());
        }

        assertEquals("The collection index file should contain all string IDs.", stringIds, actual);
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.increaseDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testIncreaseDocumentUseCountWordSourceIdSourceNotSerializable() {
        try {
            persistence
                .increaseDocumentUseCount(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testIncreaseDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_negative.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document usage count is not a number. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdUseCountNotNumber() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_notnumber.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is not a number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>increaseDocumentUseCount(WordSourceId)</code> when the document usage count is missing. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdUseCountMissing() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_missing.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.increaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>increaseDocumentUseCount(WordSourceId)</code>. The document usage count should be
     * increased by 1. This is verified by get the usage count.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testIncreaseDocumentUseCountWordSourceIdAccuracy() throws Exception {
        // Copy a valid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_normal.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Increase the usage count
        persistence.increaseDocumentUseCount(id);

        // Verify
        assertEquals("The usage count should be increased.", 6, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.decreaseDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testDecreaseDocumentUseCountWordSourceIdSourceNotSerializable() {
        try {
            persistence
                .decreaseDocumentUseCount(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testDecreaseDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_negative.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is 0. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountZero() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_zero.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is 0, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is not a number. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountNotNumber() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_notnumber.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is not a number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>decreaseDocumentUseCount(WordSourceId)</code> when the document usage count is missing. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdUseCountMissing() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_missing.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.decreaseDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>decreaseDocumentUseCount(WordSourceId)</code>. The document usage count should be
     * decreased by 1. This is verified by get the usage count.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testDecreaseDocumentUseCountWordSourceIdAccuracy() throws Exception {
        // Copy a valid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_normal.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Decrease the usage count
        persistence.decreaseDocumentUseCount(id);

        // Verify
        assertEquals("The usage count should be decreased.", 4, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document ID is <code>null</code>. Expect
     * IllegalArgumentException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdNull() throws Exception {
        try {
            persistence.getDocumentUseCount(null);
            fail("The document ID is null, should throw IllegalArgumentException.");
        } catch (IllegalArgumentException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document source identity instance is not
     * serializable. Expect IndexPersistenceException.
     */
    public void testGetDocumentUseCountWordSourceIdSourceNotSerializable() {
        try {
            persistence.getDocumentUseCount(new WordSourceId(new Object(), "class", new String[] {" "}, Locale.US));
            fail("The document source identity is not serializable, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document index is not persisted. Expect
     * IndexPersistenceException.
     */
    public void testGetDocumentUseCountWordSourceIdNotPersisted() {
        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document index is not persisted, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document usage file is invalid. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdUsageError() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_invalid.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The document usage file is invalid, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document usage count is negative. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdUseCountNegative() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_negative.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is negative, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document usage count is not a number. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdUseCountNotNumber() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_notnumber.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is not a number, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getDocumentUseCount(WordSourceId)</code> when the document usage count is missing. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdUseCountMissing() throws Exception {
        // Copy an invalid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_missing.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        try {
            persistence.getDocumentUseCount(new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US));
            fail("The use count is missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getDocumentUseCount(WordSourceId)</code>. The retrieved use count should be 5, as
     * stored in the document usage file.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetDocumentUseCountWordSourceIdAccuracy() throws Exception {
        // Copy a valid document usage file
        TestHelper.copy(new File("test_files/XmlReference/document_usage_normal.xml"), new File(
            "test_files/Xml/document_usage.xml"));

        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Verify
        assertEquals("The usage count should be 5.", 5, persistence.getDocumentUseCount(id));
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when an attribute of a document index in the collection is missing.
     * Expect IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsAttributeMissing() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_noattr.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getIndexedDocuments();
            fail("The document index file has attribute missing, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when a document index's delimiter is not escaped properly. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsDelimiterError() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_delimitererror.xml"), new File(BASE_DIR,
            "doc_index_" + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getIndexedDocuments();
            fail("The delimiter is not escaped properly, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when a document source identity cannot be deserialized. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsDeserializeError() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_deserializeerror.xml"), new File(BASE_DIR,
            "doc_index_" + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getIndexedDocuments();
            fail("The document source identity cannot be deserialized, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests <code>getIndexedDocuments()</code> when a document ID cannot be created using the data. Expect
     * IndexPersistenceException.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsIllegalID() throws Exception {
        // Create a document id
        WordSourceId id = new WordSourceId("id", "class", new String[] {" ", "\0"}, Locale.US);

        // Copy the corrupted document index file.
        TestHelper.copy(new File("test_files/XmlReference/doc_index_iderror.xml"), new File(BASE_DIR, "doc_index_"
            + IndexPersistenceHelper.generateId(id) + "_1.xml"));

        try {
            persistence.getIndexedDocuments();
            fail("The document ID cannot be created using the data, should throw IndexPersistenceException.");
        } catch (IndexPersistenceException e) {
            // good
        }
    }

    /**
     * Tests accuracy of <code>getIndexedDocuments()</code>. All persisted document index IDs should be retrieved.
     * This is verified by comparing to a reference set of IDs.
     *
     * @throws Exception pass any unexpected exception to JUnit.
     */
    public void testGetIndexedDocumentsAccuracy() throws Exception {
        Set sourceIds = new HashSet();

        // Copy the persisted document index files.
        for (int i = 0; i < 5; ++i) {
            WordSourceId id = new WordSourceId("source" + i, "class" + i, new String[] {"delimiter" + i}, Locale.US);
            String stringId = IndexPersistenceHelper.generateId(id);

            sourceIds.add(id);

            // Copy the document index persistence file
            TestHelper.copy(new File("test_files/XmlReference/doc_index_" + stringId + "_1.xml"), new File(BASE_DIR,
                "doc_index_" + stringId + "_1.xml"));
        }

        // Retrieve the persisted document indices
        Set actual = persistence.getIndexedDocuments();

        // Verify
        assertEquals("All persisted document indices should be retrieved.", sourceIds, actual);
    }
}








